
//------------------------------------------------------------------------------
// Projektname:		
// Dateiname:		$RCSfile: CStrassennetz.h,v $
// Beschreibung:	Jrg Draeger
// Version:	        $Revision: 1.123 $
// Gendert von:	$Author: pompl $
// Gendert am:		$Date: 2001/07/31 07:53:16 $
// Zustand:	        $State: Exp $
//------------------------------------------------------------------------------


#ifndef CSTRASSENNETZ_H
#define CSTRASSENNETZ_H

#include <LEDA/graph.h>
#include <LEDA/list.h>
#include <LEDA/matrix.h>
#include <LEDA/tuple.h>
#include <LEDA/rat_geo_alg.h>
#include <LEDA/graph_misc.h>

#include "../Datenstruktur/CWegenetzKnoten.h"
#include "../Datenstruktur/CWegenetzKante.h"
#include "../Datenstruktur/CPunkt.h"
#include "../Programm/CHSSchnittstelle.h"
#include "../Datenstruktur/CParameterHaltestelle.h"
#include "../Datenstruktur/CGuetemasse.h"
#include "../Datenstruktur/Debug.h"
#include "../Datenstruktur/define.h"
#include "CVisualisierung.h"
#include "CFortschritt.h"
#include "CVerbindung.h"

class CHaltestelle;

/**
 * Dies ist eine Musterklasse zur Implementierung diverser Algorithmen
 * zur Berechnung von Haltestellen.
 * Soll ein spezieller Algorithmus implementiert werden, so wird
 * hierzu eine weitere Klasse von dieser abgeleitet.
 * Somit enthaelt diese Klasse nur Methoden, welche allgemeiner
 * Natur sind.
 * In der Basisklasse werden hierbei nur die Pflichthaltestellen als
 * Ergebnis der Berechnung zurueckgegeben. Sie wird im Allgemeinen,
 * nicht instanziert. Hierfuer sind die von ihr abgeleiteten Klassen
 * vorgesehen.
 */
class CStrassennetz {
public:

  /**
   * Konstruktor.
   * Dieser wird im Allgemeinen nicht verwendet, da ansonsten
   * keine Ueberpruefung eines Abbruch-Signales von aussen stattfindet.
   * Kann dazu verwendet werden, die Berechnung der Haltestellen
   * in einem Batch-Betrieb ohne graphische Ausgabe zu erzeugen.
   * @see initialisieren(void)
   * @see Datenstruktur.CParameterHaltestelle
   */
  CStrassennetz();

  /**
   * Konstruktor.
   * Dies ist der Standard-Konstruktor, mit welchem aus dem Modul
   * Programm eine Instanz von CStrassennetz oder auch einer abgeleiteten
   * Klasse gebildet wird.
   * Ist fuer das initialisieren der relevanten Dtaenstrukturen zustaendig.
   * @param abbruch Der Zeiger auf das Flag, welches den Abbruch signalisiert.
   * @param lnkSchnittstelle Zeiger auf die Schnittstelle aus dem Programm-Modul.
   * @param Parameter Zeiger auf das Objekt, welches alle Eingabedaten kapselt.
   * @see initialisieren(void)
   * @see Datenstruktur.CParameterHaltestelle
   */
  CStrassennetz(bool * abbruch, CHSSchnittstelle * lnkSchnittstelle, CParameterHaltestelle * Parameter, CFortschritt * Fortschritt = NULL);

  /**
   * Destruktor.
   */
  virtual ~CStrassennetz();

  /**
   * Fhrt die Berechnungen aus. Kann in den abgeleiteten Klassen
   * erweitert werden.
   * Ansonsten werden die folgenden Schritte ausgefhrt:
   * - Haltestellen setzen,
   * - Haltestellen an das Strassennetz anpassen,
   * - Verkehrsbedarfsmatrix erzugen,
   * - Wegenetzgraph erstellen,
   * - aufrufen von CHSSchnittstelle::haltestellenErgebnisrueckgabe
   * .
   * @see Programm.CHSSchnittstelle#haltestellenErgebnisrueckgabe(GRAPH *,matrix *)
   */
  virtual void starten(void);

protected:
  /**
   * Dient dem Initialiseren aller Variablen und Zeiger. Wird direkt von
   * den Konstruktoren aufgerufen, welche danach weitere Modifikationen
   * am Zustand vornehmen.
   * Ruft CHSSchnittstelle::haltestellenWerteanforderung auf.
   * @see Programm.CHSSchnittstelle#haltestellenWerteanforderung(float &,float &,float &,float &,list *,list*> *,list *,matrix *,list *,matrix *,list *)
   */
  void initialisieren(void);

  /**
   * Berechnet die Haltestellen.
   * Diese Methode wird von den abgeleiteten Klassen ueberschrieben.
   */
  virtual void berechnenHaltestellen();
    
  /**
   * Erzeugt die Verkehrsbedarfsmatrix, die an die Schnittstelle zurckgegeben wird.
   * verwendet LH 540.
   */
  virtual void erzeugenVerkehrsbedarfsmatrix();

  /**
   * Berechnet die Entfernung zwischen zwei Punkten
   * - arbeitet auf dem Strassennetzgraph
   * - berechnet die Entfernung fuer den ganzen Graphen, fuer die 2 uebergebenen Punkte, die auf der Strasse liegen muessen
   * @preconditions zwei Punkte im Strassennetz mit Kreuzungen
   * die zwei uebergebenen Punkte muessen auf der Strasse liegen 
   * Kreuzungen muessen berechnet worden sein
   * @postconditions Entfernung zwischen den beiden Punkten
   * @param startpunkt Erster Punkt, von dem aus die Entfernung gemessen wird.
   * @param zielpunkt Zweiter Punkt, zu dem die Entfernung gemessen wird.
   */
  double berechnenEntfernungen(CPunkt * startpunkt, CPunkt * zielpunkt);

  /**
   * Ermittelt den kuerzesten Weg zwischen zwei Haltestellen
   * - arbeitet auf dem Strassennetzgraphen
   * - berechnet die kuerzesten Wege zwischen zwei Haltestellen mit Hilfe von LEDA-Algorithmen
   * Implementiert LH 900 bzw. Aufruf einer gleichwertigen Methode in LEDA.
   * @preconditions Strassennetz mit Kreuzungen wurde berechnet
   * Die zwei uebergebenen Punkte im Strassennetz auf Strassen
   * @postconditions Entfernung
   * @param radius Die Entfernung, in welcher die Wege zuruekcgegeben werden sollen.
   * @param startHaltestelle Haltestelle, von der aus der Weg gesucht werden soll.
   * @param zielHaltestelle Haltestelle, zu welcher der Weg gesucht werden soll.
   */
  list<CPunkt *> * ermittelnKuerzesteWege(double radius, CPunkt * startHaltestelle, CPunkt * zielHaltestelle);

  /*
   * rechnet eine gegebene Entfernung von Metern nach dem in GRASS
   * verwendeten Entfernungsmass um.
   */
  double umrechnenMeterGRASS(double meter);

  /*
   * Rechnet ein in GRASS verwendete Entfernungsmass in Metern um.
   */
  double umrechnenGRASSMeter(double grass);

  /*
   * Liefert die Luftlinienentfernung zweier Punkte.
   */
  double ermittelnEntfernung(CPunkt * startpunkt, CPunkt * zielpunkt);

  /*
   * Liefert die Gesamtlaenge eines gegebenen Weges.
   */
  double ermittelnLaenge(list<CPunkt*> * Weg, unsigned int  bisZuPosition = 0);

  /**
   * Legt die Haltestelle an den naechsten Strassenzug. Hierzu wird das Lot benutzt.
   * Implmetiert LH 500.
   * @preconditions Haltestelle darf auch auf dem Strassennetz liegen!
   * @postconditions Haltestelle wurd am Strassenentz liegend auf den naechsten Strassenzug gelegt.
   * @param haltestelle Haltestelle, von der aus auf den naechsten Strassenzug das Lot gefaellt werden soll.
   */
  CPunkt * faellenLot(CPunkt * haltestelle, bool NotfallsEndpunkteNehmen = false);

  /**
   * Faellt das Lot von der angegebenen Haltestelle auf die
   * gegebene Linie.
   *
   * @return Der CPunkt, welcher den Lotfusspunkt beschreibt oder NULL, falls dieser auf dieser nicht auf der Linie liegt.
   */
  CPunkt * faellenLot(CPunkt * haltestelle, CPunkt * linienanfang, CPunkt * linienende);

  /**
   * Pass eine gegebene Haltestelle an das Strassennetz an.
   * Moegliche Implemntierungen sind LH 500, LH 510 und LH 520 bzw. Aufruf
   * von faellenLot(), usw.
   * @postconditions Haltestelle ist an das Strassennetz angepasst.
   * @param haltestelle Haltestelle, welche an das Strassennetz angepasst werden soll.
   */
  CPunkt * anpassenHaltestellenStrassennetz(CPunkt * haltestelle);

  /**
   * Passt alle vorhandenen Haltestellen an das Strassennetz an.
   * Ruft iterativ fuer jede Haltestelle CStrassennetz::anpassenHaltestellenStrassennetz(CPunkt * haltestelle) auf.
   * @postconditions Haltestellen sind an das Strassennetz angepasst.
   * @return Liefert eine Liste mit den angepassten Haltestellen zurueck.
   * @see Haltestelle.CStrassennetz#anpassenHaltestellenStrassennetz(CPunkt *)
   */
  list<CWegenetzKnoten*> * anpassenHaltestellenStrassennetz(list<CWegenetzKnoten *> * haltestellen);

  /**
   * Passt alle vorhandenen Haltestellen an das Strassennetz an.
   * Ruft iterativ fuer jede Haltestelle CStrassennetz::anpassenHaltestellenStrassennetz(CPunkt * haltestelle) auf.
   *
   * Zudem werden alle Haltestellen entfernt, welche nach dem Ummappen zu nah
   * (Entfernung < minimalesEinzugsgebiet) liegen. Hierbei wird die zuerst
   * aufgefundene Haltestelle beibehalten.
   * @postconditions Haltestellen sind an das Strassennetz angepasst.
   * @return Liefert eine Liste mit den angepassten Haltestellen zurueck.
   * @see Haltestelle.CStrassennetz#anpassenHaltestellenStrassennetz(CPunkt *)
   */
  void anpassenHaltestellenStrassennetz();

  void bearbeitenUeberfluessigerHaltestellen(void);

protected:

  CParameterHaltestelle * Parameter;

  matrix * Verkehrsbedarfsmatrix;

  /**
   * Dem Reza sein Wegenetzgraph. 
   * 
   * Die Knoten des Wegenetzgraphen stellen die Haltestellen dar. Sie haben als Attribute eine ID, eine Prioritaet, die im Bereich von 1 bis 10 liegt und die Wichtigkeit einer Haltestelle angeben, sowie eine Position im Koordinatensystem des GIS.
   * 
   * Die Kanten habe lediglich das Attribut Entfernung, das die Entfernung zweier Haltestellen in Metern angibt. Sie stellen die Verbindung zwischen den Haltestellen ohne festen Bezug zum GIS oder Strassennetz dar. 
   */
  GRAPH<CWegenetzKnoten *, CWegenetzKante * > * Wegenetzgraph;

  /**
   * Stellt einen Zeiger auf eine LEDA-Liste dar, die WegenetzKnoten als Elemente enthaelt.
   * Dies sind die berechnenten Haltestellen, sowie die Pflichthaltestellen.
   * Zudem sind ID und Prioritaet vergeben.
   */
  list<CWegenetzKnoten*> * HaltestellenKnoten;

  /**
   * Stellt einen Zeiger auf eine LEDA-Liste dar, die Punkte als Elemente enthaelt. Dies sind die berechnenten Haltestellen, sowie die Pflichthaltestellen.
   */
//  list<CPunkt*> * Haltestellen;

  /**
   * Beinhaltet die Liste der in das Strassennetz eingefuegten Haltestellen.
   * Diese sind dabei als node vermerkt, um sie im Strassennetz wiederzufinden.
   */
  list<node*> * HaltestellenImStrassennetz;

  /**
   * Das Strassennetz ist ein Graph und stellt das eigentliche Strassennetz dar. Es wird durch die Methode berechnenKreuzungen() aufbebaut. Die Knoten stellen die Kreuzungen dar, die Kanten die Strassen zwischen den Kreuzungen
   */
  GRAPH<rat_point,rat_segment>* Strassennetz;

  /**
   * Das Attribut Fahrtwunschliste dient dazu die Eintraege der Fahrtwunschmatrix einem Punkt zuzuordnen. 
   */
  list<two_tuple<CPunkt*,int> > * Fahrtwunschliste;

  /**
   * Verweist auf das Attribut "abbrechen" der Klasse "CHaltestellte".
   * Dient somit dem Pruefen auf einen Abbruchbefehl.
   * @see Haltestelle.CHaltestelle
   */
  bool * abbrechen;

  /**
   * Begrenzt das zu betrachtende Planquadrat nach links(=Westen).
   */
  double begrenzungLinks;

  /**
   * Begrenzt das zu betrachtende Planquadrat nach rechts(=Osten).
   */
  double begrenzungRechts;

  /**
   * Begrenzt das zu betrachtende Planquadrat nach oben(=Norden).
   */
  double begrenzungOben;

  /**
   * Begrenzt das zu betrachtende Planquadrat nach unten(=Sueden).
   */
  double begrenzungUnten;

  /**
   * Verweist auf die Schnittstellenklasse.
   * Wird von der aufrufenden Klasse uebergeben und auch von dieser
   * vernichtet.
   */
  CHSSchnittstelle * lnkSchnittstelle;

  /**
   * Dient der Umrechnung von Meter nach GRASS. Die Zahl kennzeichnet dabei
   * wieviel GRASS-Einheiten einem Meter entsprechen.
   */
  static const double UmrechnungMeterGRASS = 1.0;

  /**
   * Dient der Umrechnung der Entfernung (Luftlinie) zur Annaeherung an
   * die Entfernung unter Beruecksichtigung des Realweges.
   *
   * Entspringt einem Erfahrungswert der HCR.
   */
  static const double EntfernungsFaktor = 0.6;

  /**
   * Enthaelt das Bevoelkerungsaufkommen, wobei selbiges ueber das
   * Planquadrat gerastert ist. Die Rasterung wird duch
   * BevoelkerungsrasterEinheiten vorgegeben.
   */
  matrix * Bevoelkerungsraster;

  /**
   * Enthaelt die Flaechenversorgung, wobei selbiges ueber das
   * Planquadrat gerastert ist. Die Rasterung wird duch
   * BevoelkerungsrasterEinheiten vorgegeben.
   */
  matrix * Flaechenraster;

  /**
   * Beschreibt die Anzhal der Rasterungseinheiten fuer die Fahrtwuensche.
   */
  unsigned int FahrtwunschrasterEinheiten;
  /**
   * Enthaelt die Fahrtwuensche, wobei selbige ueber das
   * Planquadrat gerastert ist. Die Rasterung wird duch
   * FahrtwunschrasterEinheiten vorgegeben.
   */
  matrix * Fahrtwunschraster;

  /**
   * Gibt an, welche Ausdehnung ein Sektor hat. Diese Sektoren
   * dienen dabei einem dem Hashing sehr aenlichen Verfahren.
   */
  static const double SektorAusdehnung = 60;

  /**
   * Beschreibt die Anzahl der Sektoren, welche ueber das Planquadrat
   * verteilt werden. wird dynamisch zur Laufzeit unter Beruecksichtigung
   * SektorAusdehnung ermittelt.
   */
  unsigned int SektorAnzahl;

  /**
   * Repraesentiert die Liste mit den Kanten in den Sektoren.
   */
  list<edge> ** Sektor;

  /**
   * Die Kosten jeder Kante fuer den Dijkstra.
   */
  // STEFAN_HIER
  //  edge_array<int> * dijkstraKanten;
  edge_array<double> * dijkstraKanten;

  /**
   * Eine spezielle Klasse, welche das Visualisieren des
   * Strassennetzes und der Haltestellen erlaubt.
   */
  CVisualisierung * StrassenVisualisierung;

  /**
   * Eine spezielle Klasse, welche das Visualisieren der
   * Bevoelkerungsversorgung erlaubt.
   */
  CVisualisierung * BevoelkerungsversorgungVisualisierung;

  /**
   * Eine spezielle Klasse, welche das Visualisieren der
   * Flaechenversorgung erlaubt.
   */
  CVisualisierung * FlaechendeckungVisualisierung;

  /**
   * Eine spezielle Klasse, welche das Visualisieren der
   * Legende erlaubt.
   */
  CVisualisierung * LegendeVisualisierung;

  /**
   * Enthaelt das Maximum aus der Bevoelkerungsrasterung.
   */
  long double MaximumBevoelkerung;

  /**
   * Enthaelt das Maximum aus der Flaechenrasterung.
   */
  long double MaximumFlaeche;

  /**
   * Eine Instanz, welche das graphische Ausgeben von Informationen
   * zum Fortschritt der Berechnung erlaubt.
   */
  CFortschritt * Fortschritt;

  /**
   * Enthaelt die Liste mit den noch zu pruefenden Verbindungen,
   * um den Wegenetzgraph zusammenhaengend zu machen.
   */
  list<CVerbindung * > * verbindungsListe;

  /**
   * Zeiger auf die Guetemasse, welche an die Schnitstelle
   * zurueckgeliefert werden.
   * Dieses Objekt wird bei der Initialisierung erzeugt, aber nur durch
   * die Schnittstelle vernichtet!
   */
  CGuetemasse * Guetemasse;

  /**
   * Enthaelt ein Flag, ob suchenVerbindung() ueebr eine Haltestelle
   * gelaufen ist, oder nicht.
   */
  bool UeberHaltestellenGelaufen;

  /**
   * Liste der besuchten Knoten bei der Realwegumgebung einer Haltestelle
   */

  list<node*> * BesuchteKnotenListe;

  /**
   * Liste der Randknoten, die das Realwegumgebungspolygon der Haltestelle
   * bilden
   */

  // ?? //

  unsigned int MaximaleAnzahlHaltestellen;

protected:

  /**
   * Hilfsfunktion, die den Schnittpunkt zweier Geraden berechnet und zurckgibt
   * @param m1 Steigungskoeffizient der ersten Geraden.
   * @param t1 Y-Versatz der ersten Geraden.
   * @param m2 Steigungskoeffizient der zweiten Geraden.
   * @param t2 Y-Versatz der zweiten Geraden.
   */
  CPunkt* berechnenSchnittpunkt(double m1, double t1, double m2, double t2);

  /**
   * Hilfsfunktion, die einen Punkt einer Geraden berechnet und zurckgibt
   * @param m Steigungskoeffizient.
   * @param t Y-Versatz.
   * @param x Wert.
   */
  CPunkt* berechnenPunkt(double m, double t, double x);
    
  /**
   * Erstellt den StrassennetzGraph. Ruft hierzu die Methode
   * LEDA::SWEEP_SEGMENTS() auf.
   */
  void erzeugenStrassennetzGraph();

 /**
  *  Erzeugen einer Ausgabe des StrassennetzGraphen
  */
 void ausgebenStrassennetzGraph();

  /**
   * Rastert die vorhandene Bevoelkerung unter Beruecksichtigung von
   * Bevoelkerungsraster und BevoelkerungsrasterEinheiten.
   * @precondition Bevoelkerungsraster ist initialisert und erzeugt.
   * @see Bevoelkerungsraster
   * @see BevoelkerungsrasterEinheiten.
   */
  double * rasternBevoelkerung(list<CBevoelkerungsdichteElement *> * Bevoelkerungsdichtenliste);

  /**
   * Rastert die vorhandene Flaeche unter Beruecksichtigung von
   * Bevoelkerungsraster und BevoelkerungsrasterEinheiten.
   * @precondition Bevoelkerungsraster ist initialisert und erzeugt.
   * @see Flaechenraster
   * @see BevoelkerungsrasterEinheiten.
   */
  double * rasternFlaeche(void);

  /**
   * Rastert die vorhandenen Fahrtwuenche unter Beruecksichtigung von
   * Fahrtwunschsraster und FahrtwunschrasterEinheiten.
   * @precondition Fahrtwunschraster ist initialisert und erzeugt.
   * @see Fahrtwunschraster
   * @see FahrtwunschrasterEinheiten.
   */
  void rasternFahrtwuensche(void);

  /**
   * Berechnet dem Reza sein Wegenetzgraph.
   * Dabei werden zunaechst aus dem Strassennetz alle Kreuzungen und Start- und Endpunkte von Strassenzuegen
   * entfernt. Somit bleiben zum Schluss nur noch Haltestellen als Knoten im Graphen uebrig. Beim entfernen von 
   * Kreuzungen werden Kanten doppelt auftreten. Sobald eine Kante zwischen zwei Knoten eingefuegt werden soll,
   * wo bereits eine Kante existiert, wird ihre Gesamtlaenge entlang der realen Strassen (!) verglichen und nur
   * die kuerzere Kante gespeichert, die andere wird verworfen (geloescht).
   * @precondition Strassennetz existiert komplett, Haltestellen sind gesetzt.
   */

  GRAPH<CWegenetzKnoten *, CWegenetzKante * > * berechnenWegenetzgraph(GRAPH<rat_point,rat_segment> *Strassennetz, list<CWegenetzKnoten*> *Haltestellen);

  /**
   * Ruft die parametrisierte Version von berechnenWegenetzgraph() auf
   * und stellt das Ergebnis in Wegenetzgraph zur Verfuegung.
   */
  void berechnenWegenetzgraph(void);

  /**
   * Gibt eine Berwertung der uebergebenen Haltestellen zurueck.
   * Dabei werden beruecksichtigt:
   * (Bevoelkerungsabdeckung)
   * (Flaechenabdeckung)
   * (Fahrtwunschabdeckung)
   * Doppelte Abdeckungen werden als Strafe hinzugezaehlt.
   */
  double bewerten(list<CWegenetzKnoten*> * HaltestellenKnoten);

  /**
   * Bestimmt die Ausdehnung des uebergebenen
   * Strassennetzes und traegt die vorgefundenen
   * Werte als Begrenzungen ein.
   *
   * @param Strassennetzliste Eine Liste von Listen aus Punkten, welche das Strassennetz repraesentiert.
   */
  void suchenBegrenzungen(list<list<CPunkt*>* > * Strassennetzliste);

  /**
   * Gibt die zugehoerigen Knoten zur uebergebenen Kante zurueck.
   */
  void holeKnoten(edge* AktuelleKante, CPunkt* Start, CPunkt* Ziel);

  /**
   * Sucht auf dem strassennetz die kuerzeste Verbindung
   * vom StartPunkt zum EndPunkt. Wird dabei ueber
   * eine Haltestelle gelaufen, so wird kein Pfad zurueckgeben,
   * ansonsten enthaelt der Graph eine Liste der Punkte,
   * welchen den abzufahrenden Linienzug einschliesslich
   * StartPunkt und EndPunkt, enthaelt.
   *
   * @param StartPunkt Punkt, von welchem die Verbindung aufgebaut werden soll.
   * @param EndPunkt Punkt, zu welchem die Verbindung aufgebaut werden soll.
   * @param AufHaltestellenPruefen Gibt an, ob auf Haltestellen geprueft werden soll, also ob ein Pfad auch dann gueltig ist, wenn er ueber eine Haltestelle laeuft.
   * @returns Liste mit dem Linienzug oder NULL.
   */
  list<CPunkt *> * suchenVerbindung(CPunkt * StartPunkt, CPunkt * EndPunkt, bool AufHaltestellenPruefen = true);

  /**
   * Sucht auf dem strassennetz die kuerzeste Verbindung
   * vom StartPunkt zum EndPunkt. Wird dabei ueber
   * eine Haltestelle gelaufen, so wird kein Pfad zurueckgeben,
   * ansonsten enthaelt der Graph eine Liste der Punkte,
   * welchen den abzufahrenden Linienzug einschliesslich
   * StartPunkt und EndPunkt, enthaelt.
   *
   * @param StartPunkt Punkt, von welchem die Verbindung aufgebaut werden soll.
   * @param EndPunkt Punkt, zu welchem die Verbindung aufgebaut werden soll.
   * @param AufHaltestellenPruefen Gibt an, ob auf Haltestellen geprueft werden soll, also ob ein Pfad auch dann gueltig ist, wenn er ueber eine Haltestelle laeuft.
   * @returns Liste mit dem Linienzug oder NULL.
   */
  list<CPunkt *> * suchenVerbindung(node * Ziel, node * Start, bool AufHaltestellenPruefen = true);

  /**
   * Gibt den WegenetzGraph aus.
   *
   * @preconditions Wird nur ausgefhrt, wenn DEBUG definiert ist.
   */
  void ausgebenWegenetzGraph(GRAPH<CWegenetzKnoten *, CWegenetzKante *> * Wegenetzgraph);

  /**
   * Erstellt aus dem unter Umstaenden noch unzusammenhaengenden
   * WegenetzGraphen einen zusammenhaengenden. Hierzu wird eine
   * Variante von Kruskal vewendet, wobei davon ausgegangen wird,
   * dass bis zum Verbindungsradius schon entsprechend vorgearbeitet
   * wurde und nur noch Verbindungen groesser als der Verbindungsradius
   * geprueft werden muessen.
   * Als Vorgabe dient dabei die verbindungsListe.
   * @see sortierenVerbindungsListe(void)
   * @see berechnenWegenetzgraph(void)
   */
  void erstellenZusammenhangWegenetzgraph(void);

 /**
  * Nummeriert die vorhandenen Haltestellen mit Null beginnend durch.
  *
  * @postonditions Alle Haltestellen aus HaltestellenKnoten sind mit 0 beginnend fortlaufend durchnummeriert.
  */
  void nummerierenHaltestellen(void);

 /**
  * Ermittelt die Knoten rechts und links von einer Kante, die Kreuzungen darstellen.
  * Knoten, die keine Kreuzungen sind werden ignoriert
  *
  * @param Kante
  * @preconditions Alle Knoten haben das Attribut Haltestelle, Kreuzung, Strassenpunkt
  */
  void ermittelnInzidenteKnoten(edge * Kante, node *& linkerKnoten, node *& rechterKnoten);

  /**
  * Ermittelt die Knoten rechts und links von einer Kante, die Kreuzungen darstellen.
  * Knoten, die keine Kreuzungen sind werden in eine Liste eingetragen und zurueckgegeben
  *
  * @param Kante
  * @preconditions Alle Knoten haben das Attribut Haltestelle, Kreuzung, Strassenpunkt
  */
  void ermittelnInzidenteKnoten(edge * Kante, node * linkerKnoten, node * rechterKnoten, list<node *> * linkeKnotenfolge, list<node *> * rechteKnotenfolge);


  /**
   * Baut die Sektoren auf, welche die Kanten des Strassennetzes
   * enthalten sollen. Die Idee eines Sektors ist dabei einem
   * Quadtree nicht unaehnlich:
   * Man teilt das gesamte Planquadrat in mehrere Einehiten, welche
   * quadratisch angeordnet sind. Dadurch laesst sich fuer jeden
   * Punkt eindeutig entscheiden, in welchem Sektor er liegt.
   */
  void aufbauenSektoren(void);

  /**
   * Prueft den Sektor, in welchem sich ein Punkt befindet.
   */
  int pruefenSektor(rat_point p);

  /**
   * Fuegt eine Linie in die beruehrten Sektoren ein.
   * Verwendet pruefenSektor(rat_point).
   */
  void einfuegenStrassensegmentSektor(edge Kante);

  /**
   * Liefert alle Strassen aus dem uebergebenen Sektor plus den
   * direkt angrenzenden Sektoren.
   * Dabei werden niemals Kanten doppelt ausgegeben.
   *
   * @param Sektor Sektor, welcher der Ursprung der Ausgabe ist.
   *
   * @return Eine Liste alle Kanten, welche in dem Sektor und aus den umliegenden ermittelt wurden, ohne doppelte Kanten.
   */
  list<edge> * sammelnStrassenAusSektoren(int Sektor);

  /**
   * Gibt die Liste mit den Haltestellen aus.
   */
  void ausgebenHaltestellen();

  /**
   * Fuegt die Haltestellen in das Strassennetz ein. Dient als Grundlage fuer
   * die Erstellung des Wegenetzgraphen.
   * Dabei werden die Haltestellen als Knoten mit negativen Vorzeichen
   * behandelt, um sie spaeter wieder erkennen zu koennen.
   *
   * @param Strassennetz Das Strassennetz, in welches die Haltestellen eingefuegt werden sollen.
   * @param Haltestellen Die Haltestellen, welche eingefuegt werden sollen.
   *
   * @return Die Liste der neu erzeugten Knoten, in der gleichen Reihenfolge, wie die zugehoerigen Haltestellen.
   */
  list<node*> * einfuegenHaltestellenStrassennetz(GRAPH<rat_point,rat_segment> * Strassennetz, list<CWegenetzKnoten*> * Haltestellen);

  /**
   * Kuerzt die Koordinaten der Knoten (=Haltestellen) im uebergebenen
   * Graphen auf zwei Nachkommastellen.
   */
  void kuerzenKoordinaten(GRAPH<CWegenetzKnoten*,CWegenetzKante*> * Wegenetzgraph);

  /**
   * Fordert entsprechend der Ausdehnung des Planquadrates eine bestimmte
   * Anzahl an Sektoren an.
   * Dabei wird eine Sektor mit 'Ausdehnung'^2 angenommen.
   *
   * @postcondition Wenigstens ein Sektor wurde angefordert.
   */
  void anfordernSektoren(double Ausdehnung);

  /**
   * Zur Realwegradiusberechnung um eine Haltestelle herum
   */
  void berechnenRealwegumgebung(GRAPH<rat_point,rat_segment> *, node *);

  double testobKreuzung(GRAPH<rat_point,rat_segment> *, node *, double);
  
  double pruefeLaenge(GRAPH<rat_point,rat_segment> *, node *, node *, double);

  /**
   * Sortiert die verbindungsListe, bestehend aus Zeigern auf Objekten
   * des Typs CVerbindung, aufsteigend nach ihrer Laenge.
   * Wird zur Vorbereitung von erstellenZusammenhangWegenetzgraph()
   * benoetigt.
   * @see erstellenZusammenhangWegenetzgraph(void)
   */
  void sortierenVerbindungsListe(void);

  /**
   * Entfernt aus dem Strassennetz lose Teile.
   * Hierzu werden alle Komponenten berechnet und dann alle Teile,
   * welche nicht zur groessten gehoeren entfernt.
   */
  void kuerzenStrassennetz(void);

  /**
   * Dies ist ein Koeffizient, welcher dazu dient, die Anpassung der Haltestellen
   * zusaetzlich von der Entfernung der zuerst berechneten Haltestellenlage
   */
  double berechnenModifizierterKoeffizientHaltestellenLage(double Anteil);
};
#endif //CSTRASSENNETZ_H
